#include <stdint.h>
#include <string.h>
#include <inttypes.h>
#include <openssl/evp.h>
#include "common.h"
#include "xor.h"

Result_t ecb_decrypt(
                    CryptoAttribs_t* attribs,
                    const Key_t* const key)
{
    // if output is NULL then allocate same size as input rounded to block size
    if(attribs->output == NULL)
    {
        attribs->output = (uint8_t*) malloc(
                    ROUNDUP_16_BYTES(attribs->input_len) * sizeof(uint8_t));
    }

    int ret, ret_fin;
    ret=ret_fin=0;
    EVP_CIPHER_CTX ctx;
    EVP_CIPHER_CTX_init(&ctx);

    EVP_CipherInit_ex(&ctx, EVP_aes_128_ecb(), NULL, key->key, NULL, 0);
    EVP_CIPHER_CTX_set_key_length(&ctx, key->len);

    EVP_DecryptUpdate(&ctx, attribs->output, &ret, attribs->input, attribs->input_len);
    EVP_DecryptFinal (&ctx, (attribs->output)+ret, (int*)&attribs->output_len);
    return Result_OK;
}

Result_t ecb_encrypt(
                    CryptoAttribs_t* attribs,
                    const Key_t* const key)

{
    // if output is NULL then allocate same size as input rounded to block size
    if(attribs->output == NULL)
    {
        attribs->output =  (uint8_t*) malloc(
                    ROUNDUP_16_BYTES(attribs->input_len) * sizeof(uint8_t));
    }

    int ret, ret_fin;
    ret=ret_fin=0;
    EVP_CIPHER_CTX ctx;
    EVP_CIPHER_CTX_init(&ctx);

    EVP_EncryptInit  (&ctx, EVP_aes_128_ecb(), key->key, NULL);
    EVP_CIPHER_CTX_set_key_length(&ctx, key->len);

    EVP_EncryptUpdate(&ctx, attribs->output, &ret, attribs->input, attribs->input_len);
    EVP_EncryptFinal (&ctx, (attribs->output)+ret, (int*)&attribs->output_len);
    return Result_OK;
}


Result_t cbc_decrypt(
            CryptoAttribs_t* attribs,
            const Key_t* const key)
{
    if((attribs->input_len % key->len) != 0 )
        return Result_Error;
    const size_t bc = attribs->input_len / key->len;
    if( attribs->output == NULL)
    {
        // same as input + 16 bytes iv
        const size_t len = (ROUNDUP_16_BYTES(attribs->input_len)+16);
        attribs->output =  (uint8_t*) malloc(len * sizeof(uint8_t));
        attribs->output_len = len;
    }

    uint8_t* iv = attribs->iv;
    CryptoAttribs_t ecb_attribs;
    CryptoAttribs_t::Init(&ecb_attribs);
    ecb_attribs.output = (uint8_t*) malloc(key->len);
    ecb_attribs.output_len = key->len;

    for(int i=0; i<bc; ++i)
    {
        ecb_attribs.input = (attribs->input)+(key->len*i);
        ecb_attribs.input_len = key->len;

        // 1. Decrypt
        if( ecb_decrypt(&ecb_attribs, key) != Result_OK )
            return Result_Error;

        // 2. Xor IV
        xor_repeatedly(
            iv, attribs->iv_len,
            ecb_attribs.output, ecb_attribs.output_len,
            (attribs->output)+(key->len*i));

        printf("%.16s\n", attribs->output);
        // 3. Change iv with block
        iv = ecb_attribs.input;
        attribs->output_len += key->len;
    }
    free(ecb_attribs.output);

    return Result_OK;
}

Result_t cbc_encrypt(
                    CryptoAttribs_t* attribs,
                    const Key_t* const key)
{
    if((attribs->input_len % key->len) != 0 )
        return Result_Error;

    if( attribs->iv_len != key->len )
        return Result_Error;

    const size_t bc = attribs->input_len / key->len;

    if( attribs->output == NULL)
    {
        // same as input + 16 bytes iv
        attribs->output =  (uint8_t*) malloc(
                    (ROUNDUP_16_BYTES(attribs->input_len)+16) * sizeof(uint8_t));
    }

    uint8_t* iv = attribs->iv;

    CryptoAttribs_t ecb_attribs;
    CryptoAttribs_t::Init(&ecb_attribs);
    ecb_attribs.input = (uint8_t*) malloc(key->len);
    ecb_attribs.input_len = key->len;
    ecb_attribs.output = (uint8_t*) malloc(key->len);
    ecb_attribs.output_len = key->len;
    for(int i=0; i<bc; ++i)
    {
        uint8_t* i_buff_ptr = (attribs->input)+(key->len*i);

        // 1. Xor IV
        xor_repeatedly(
            iv, attribs->iv_len,
            i_buff_ptr, attribs->iv_len,
            ecb_attribs.input);
        ecb_attribs.input_len = key->len;

        // 2. Encrypt
        if( ecb_encrypt(&ecb_attribs, key) != Result_OK )
            return Result_Error;

        // 3. Change iv with block
        iv = i_buff_ptr;
        memcpy((attribs->output)+(key->len*i),
                ecb_attribs.output,
                ecb_attribs.output_len);
        attribs->output_len += ecb_attribs.output_len;
    }
    free(ecb_attribs.input);
    free(ecb_attribs.output);

    return Result_OK;
}